/*
** Module   : NativeAPIEntry.java
** Abstract : stores metadata relating to a native API call 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20131125 First version.
*/

package com.goldencode.p2j.library;

import com.goldencode.p2j.util.*;

/**
 * Container for legacy info belonging to an native API call.
 */
public class NativeAPIEntry
extends InternalEntry
{
   /** Library name in which this native API exists. */
   private String libname = null;
   
   /** The API calling convention. */
   private CallingConvention callconv = CallingConvention.DEFAULT;
   
   /** Flag to determine if a library loaded just for this call should be kept in memory. */
   private boolean persistent = false;
   
   /** The ordinal if the API is exported via ordinal instead of name. */
   private int ordinal = -1;
   
   /**
    * Construct an instance.
    * 
    * @param    pname
    *           Legacy 4GL name for this internal-entry.
    * @param    jname
    *           Converted Java name for this internal-entry.
    * @param    type
    *           The type of this internal entry.
    */
   public NativeAPIEntry(String pname, String jname, String type)
   {
      super(pname, jname, type);
   }
   
   /**
    * Set the value of the given attribute.  Any attributes that are specific to the native
    * API call will be parsed and stored as members, instead of being added to the map of
    * attributes.
    * 
    * @param    name
    *           The attribute's name.
    * @param    value
    *           The attribute's value.
    */
   public void putAttribute(String name, String value)
   {
      // TODO: we could also check for empty string here, but I don't know if it is safe (it
      //       might be used for the non-native attributes)
      if (name == null || value == null)
         return;
      
      name = name.toLowerCase();
      
      if ("libname".equals(name))
      {
         libname = value;
      }
      else if ("callconv".equals(name))
      {
         try
         {
            callconv = CallingConvention.valueOf(value.toUpperCase());
         }
         
         catch (IllegalArgumentException iae)
         {
            // misformed value, ignore it
         }
      }
      else if ("persistent".equals(name))
      {
         // any misformed value should parse as false
         persistent = Boolean.parseBoolean(value);
      }
      else if ("ordinal".equals(name))
      {
         try
         {
            ordinal = Integer.parseInt(value);
         }
         
         catch (NumberFormatException nfe)
         {
            // misformed value, ignore it
         }
      }
      else
      {
         super.putAttribute(name, value);
      }
   }

   /**
    * Get the library name.
    *
    * @return   The library name.
    */
   public String getLibraryName()
   {
      return libname;
   }
   
   /**
    * Get the calling convention.
    *
    * @return   The calling convention.
    */
   public CallingConvention getCallingConvention()
   {
      return callconv;
   }
   
   /**
    * Checks if the library should be loaded persistently.
    *
    * @return   <code>true</code> if the library should be loaded persistently.
    */
   public boolean isPersistent()
   {
      return persistent;
   }
   
   /**
    * Get the ordinal.
    *
    * @return   The ordinal that the API is exported as or -1 if no ordinal was specified.
    */
   public int getOrdinal()
   {
      return ordinal;
   }
}
